/*
	Dots CustomDataType & CustomGui Example
	(C) 2021 MAXON Computer GmbH

	Author: Sebastian Bach, Ferdinand Hoppe
	Date: 13/12/2021

	Showcases how to use custom data types in combination with custom GUIs.

  The type iSplineVisDataType is the actual data object that stores the custom data, where
	SplineVisDataType is the plugin class that represents the new data type within Cinema 4D, 
	handling the underlying iSplineVisDataType data.

  SplineVisCustomGUI is the plugin class that represents a new GUI element for that data
  type. It creates an SplineVisGUI dialog that realizes the GUI for that data type. The
  dialog includes the GeUserArea implementation SplineVisUserArea which handles the rendering of the
  custom data in a GUI.
*/

#include "c4d.h"
#include "c4d_gui.h"
// #include "c4d_symbols.h"
#include "splinevis.h"
#include "lib_prefs.h"

#include "maxon/dataserialize.h"
#include "maxon/apibase.h"
//#include <chrono>



#define ID_SETTINGS_DIALOG 10269420



//---------------------------
/// The user area used to display the custom datatype
//---------------------------
SplineVisUserArea::SplineVisUserArea() : _splvisData(new iSplineVisDataType)
{
	m_animating = true;
	m_animationTime = 0.0f; // Initialize animation time
	//m_animationDuration = 5.0f; // Set the duration of the
	m_animationDuration = KNOB_AN_DURATION ; // Set the duration of the

	_settingsDiag.grad = _splvisData->_gradient;
}

SplineVisUserArea::~SplineVisUserArea()
{
	//StopAnimation();
	m_animating = false;
}

Bool SplineVisUserArea::Init()
{
	SetTimer(KNOB_REFRESH_INTERVAL); // Set timer interval (in milliseconds)
	return true;
}

Bool SplineVisUserArea::InitValues()
{
	return true;
}

Bool SplineVisUserArea::GetMinSize(Int32& w, Int32& h)
{
	w = 200;
	h = 50;
	return true;
}

void SplineVisUserArea::DrawMsg(Int32 x1, Int32 y1, Int32 x2, Int32 y2, const BaseContainer& msg)
{
	this->OffScreenOn();

	// drawing the white background
	Int32 width = GetWidth();
	Int32 height = GetHeight();
	
	this->DrawSetPen(GetViewColor(PREFS_SCHEMECOLORS));
	this->DrawRectangle(0, 0, width, height);
	
	// Sliders Deck	
	this->DrawSetPen(Vector(0.08));
	this->DrawRectangle(height/2, height * 0.4, width-(height / 2), height * 0.6);
	//		chmfered deck
	this->DrawEllipseFill(Vector2d(height / 2, height * 0.5), Vector2d(height * 0.2 * 0.5));
	this->DrawEllipseFill(Vector2d(width - (height / 2), height * 0.5), Vector2d(height * 0.2 * 0.5));

	if (_splvisData)
	{
		// Knob Demo
		//Handlling Spline:
		Float64 pointHeight;
		if (_splvisData->_spline != nullptr)
		{
			pointHeight = _splvisData->_spline->GetPoint(m_animationTime / m_animationDuration).y;
		}

		Float max_knobSize = (height / 2) * (1 - KNOB_MARGIN);
		Float min_knobSize = max_knobSize * KNOB_MIN_SIZE;
		Float knob_path = width - (height);
		Float pureSize = max_knobSize - min_knobSize;

		Float knobSize = (pointHeight * pureSize) + min_knobSize;

		//		Knobs color based on the knob size:
		InitRenderStruct irs;
		irs.Init(GetActiveDocument());
		iferr(_splvisData->_gradient->InitRender(irs))
			return;

		const Vector knobColor = _splvisData->_gradient->CalcGradientPixel((m_animationTime / m_animationDuration));

		const Vector2d knobPos = Vector2d((knob_path * (m_animationTime / m_animationDuration)) + (height / 2), height / 2);
		_splvisData->_gradient->FreeRender();


		this->DrawSetPen(Vector(0.6));
		this->DrawEllipseFill(knobPos, Vector2d(knobSize));
		this->DrawSetPen(knobColor);
		this->DrawEllipseFill(knobPos, Vector2d(knobSize * 0.8));


		// Access the object(tag)'s container
		this->DrawSetPen(Vector(1.0, 0.5, 0.3));
		//this->DrawText("Testies"_s, width * 0.9, height * 0.6, 0);
		this->DrawText(String::FloatToString(m_animationTime), width * 0.9, height * 0.6, 0);

		// if the data is defined for each dot a rectangle is drawn
		this->DrawSetPen(Vector(0.0));

		for (Int32 i = 0; i < _splvisData->_points.GetCount(); ++i)
		{
			const Vector vec = _splvisData->_points[i];
			this->DrawRectangle(SAFEINT32(vec.x) - 5, SAFEINT32(vec.y) - 5,
				SAFEINT32(vec.x) + 5, SAFEINT32(vec.y) + 5);
		}
	}
}

Bool SplineVisUserArea::InputEvent(const BaseContainer& msg)
{
	// check the input device
	Int32 dev = msg.GetInt32(BFM_INPUT_DEVICE);
	Int32 chn = msg.GetInt32(BFM_INPUT_CHANNEL);
	switch (dev)
	{
		// some mouse interaction
		case	BFM_INPUT_MOUSE:																	
		{
			switch (chn)
			{
				case BFM_INPUT_MOUSELEFT:
				{
					// get the cursor position
					Int32 mx = msg.GetInt32(BFM_INPUT_X);
					Int32 my = msg.GetInt32(BFM_INPUT_Y);
					Global2Local(&mx, &my);

					// add the new position to the data
					iferr(_splvisData->_points.Append(Vector(mx, my, 0)))
						return false;

					// inform the parent that the data has changed
					BaseContainer m(BFM_ACTION);
					m.SetInt32(BFM_ACTION_ID, GetId());
					SendParentMessage(m);

					Bool	dc = msg.GetBool(BFM_INPUT_DOUBLECLICK);
					if (dc)
					{
						_settingsDiag.grad = _splvisData->_gradient;
						_settingsDiag.Open(DLG_TYPE::ASYNC, ID_SETTINGS_DIALOG, -2, -2, 350, 200);
						//gradient = _settingsDiag.grad;
					}


					break;
				}
				case BFM_INPUT_MOUSERIGHT:
				{
					_splvisData->_points.Flush();
					Redraw();

					break;
				}
			}

			return true;
		}
	}
	return false;
}


void SplineVisUserArea::Timer(const BaseContainer& msg)
{
	// Handle timer events
	if (m_animating)
	{
		// Calculate elapsed time since last frame



		//std::chrono::steady_clock::time_point currentTime = std::chrono::steady_clock::now();
		//delta_time = (currentTime - m_lastTime).count();


		m_animationTime++;
		if (m_animationTime > m_animationDuration) {
			// Reached the end of the loop, reset frame to start
			m_animationTime = 0;
		}


		this->Redraw();

		//m_lastTime = currentTime;
	}
}

//void SplineVisUserArea::StartAnimation()
//{
//	//m_startTime = BaseTime::Get();
//	//m_startTime = GetActiveDocument()->GetTime();
//	m_startTime = std::chrono::steady_clock::now();
//	m_lastTime = m_startTime;
//	m_animating = true;
//}
//
//void SplineVisUserArea::StopAnimation()
//{
//	m_animating = false;
//}




//------------
/// the iCustomGui dialog hosting the user area 
//------------
SplineVisGUI::SplineVisGUI(
	const BaseContainer &settings, CUSTOMGUIPLUGIN *plugin) : iCustomGui(settings, plugin)
{
	_tristate = false;
};

Bool SplineVisGUI::CreateLayout()
{
	BaseContainer bc;

	GroupBegin(1000, BFH_SCALEFIT|BFV_FIT, 1, 1, String(), 0);
	{
		GroupSpace(5, 5);

		// creating the GeUserArea
		C4DGadget* userarea = this->AddUserArea(ID_USERAREA, BFH_LEFT | BFH_SCALEFIT, 400, 50);
		this->AttachUserArea(_SplineVisUserArea, userarea);	

		// assigning the data
		_SplineVisUserArea._splvisData = &_splvisData;
	}
	GroupEnd();

	return SUPER::CreateLayout();
};

Bool SplineVisGUI::InitValues()
{
	const maxon::BaseArray<Vector> value = _splvisData.GetValue();
	_SplineVisUserArea._splvisData = &_splvisData;
	_SplineVisUserArea.Redraw();
	// _SplineVisUserArea._splvisData->SetValue(value);

	return true;
};

Bool SplineVisGUI::Command(Int32 id, const BaseContainer &msg)
{
	iferr_scope_handler
	{
		return false;
	};

	switch (id)
	{

		// A message from the user area was received, inform the parent that the data has changed.
		case ID_USERAREA:
		{
			BaseContainer m(msg);
			m.SetInt32(BFM_ACTION_ID, GetId());
			m.SetData(BFM_ACTION_VALUE, this->GetData().GetValue());
			SendParentMessage(m);

			// Redraw the user area
			_SplineVisUserArea.Redraw();
			return true;
		}
    default:
      break;
	}

	return SUPER::Command(id, msg);
}

Bool SplineVisGUI::SetData(const TriState<GeData> &tristate)
{
	// data is set from the outside

	const iSplineVisDataType* const data = static_cast<const iSplineVisDataType*>(
		tristate.GetValue().GetCustomDataType(ID_DTYPE_SPLINE_VIS));

	if (data)
	{
		_splvisData._points.Flush();
		iferr (_splvisData._points.CopyFrom(data->_points))
      return false;

		_SplineVisUserArea.Redraw();	
	}

	return true;
};

TriState<GeData> SplineVisGUI::GetData()
{
	TriState<GeData> tri;
	tri.Add(GeData(ID_DTYPE_SPLINE_VIS, _splvisData));

	return tri;
};

//// iSplineVisDataType::iSplineVisDataType(const maxon::BaseArray<Vector> &points)
//iSplineVisDataType::iSplineVisDataType(const BaseContainer &SplVisCon)
//{
//	splvisContainer.SetData(SPLINEVIS_REF_SPLINE, 0);
//	splvisContainer.SetData(SPLINEVIS_COLOR_RAMP, 0);
//	splvisContainer.SetFloat(SPLINEVIS_KNOB_SIZE, 0.8);
//	splvisContainer.SetFloat(SPLINEVIS_TIME_MULTIPLIER, 1.0);
//
//	// iferr_scope;
//	//                                                     _points.CopyFrom(points) iferr_ignore("do not return any error inside the constructor"_s);
//};

//------------
// this code is only needed if this Classic API data/UI have to be used in the node Editor and need
// to converted to a Maxon Data.



iSplineVisDataType::iSplineVisDataType()
{
	_gradient = Gradient::Alloc();
	_spline = SplineData::Alloc();

	if (_gradient != nullptr)
	{
		{
			GradientKnot knot;
			knot.pos = 0.0;
			knot.col = Vector(0, 0, 1);
			knot.bias = 0.8;
			_gradient->InsertKnot(knot);
		}
		{
			GradientKnot knot;
			knot.pos = 0.5;
			knot.col = Vector(1, 1, 1);
			knot.bias = 0.2;
			_gradient->InsertKnot(knot);
		}
		{
			GradientKnot knot;
			knot.pos = 1.0;
			knot.col = Vector(1, 0, 0);
			_gradient->InsertKnot(knot);
		}
	}
	if (_spline)
	{
		_spline->SetRange(0, 1, 0.1, -1, 1, 0.1);

		_spline->MakeLinearSplineBezier(2);  // create spline


		_spline->Flip();

		Int32 count = _spline->GetKnotCount();
		if (count == 2)
		{
			CustomSplineKnot* knot0 = _spline->GetKnot(0);
			if (knot0)
			{
				knot0->vPos = Vector(0, 1.0, 0);
				knot0->vTangentLeft = Vector(-0.3, 0, 0);
				knot0->vTangentRight = Vector(0.3, 0, 0);
			}

			CustomSplineKnot* knot2 = _spline->GetKnot(1);
			if (knot2)
			{
				knot2->vPos = Vector(1, 0, 0);
				knot2->vTangentLeft = Vector(-0.3, 0, 0);
				knot2->vTangentRight = Vector(0.3, 0, 0);
			}
		}
	}










}
iSplineVisDataType::~iSplineVisDataType()
{
	Gradient::Free(_gradient);
	SplineData::Free(_spline);
}
iSplineVisDataType::iSplineVisDataType(const iSplineVisDataType& cdtd)
{
	//splvisContainer.SetData(SPLINEVIS_REF_SPLINE, nullptr);
	//splvisContainer.SetData(SPLINEVIS_COLOR_RAMP, nullptr);
	//splvisContainer.SetFloat(SPLINEVIS_KNOB_SIZE, 0.8);
	//splvisContainer.SetFloat(SPLINEVIS_TIME_MULTIPLIER, 1.0);


																	//iferr_scope;
																	//_points.CopyFrom(cdtd._points) iferr_ignore("do not return any error inside the constructor"_s);
};

maxon::Result<void> iSplineVisDataType::DescribeIO(const maxon::DataSerializeInterface& stream)
{
	iferr_scope;

	PrepareDescribe(stream, iSplineVisDataType);
	Describe("_points", _points, Vector, maxon::DESCRIBEFLAGS::TYPE_ARRAY) iferr_return;

	return maxon::OK;
}
//------------

maxon::BaseArray<Vector> iSplineVisDataType::GetValue() const 
{
	maxon::BaseArray<Vector> returnedValue;
	iferr (returnedValue.CopyFrom(_points))
	{
		return {};
	}
	return returnedValue;
}
void iSplineVisDataType::SetValue(const maxon::BaseArray<Vector> &value)
{
	_points.Flush();
	_points.CopyFrom(value) iferr_ignore("the array is already empty."_s);
}




// Register the customDataTypeDot as a Maxon Data for the Node API example.
MAXON_DATATYPE_REGISTER(iSplineVisDataType);
//---------------------
/// custom GUI data
//---------------------
Int32 SplineVisCustomGUI::GetId()
{
	return ID_GUI_SPLINE_VIS;
};

CDialog* SplineVisCustomGUI::Alloc(const BaseContainer& settings)
{
	// creating the dialog that "is" the actual GUI

	iferr (SplineVisGUI* const dlg = NewObj(SplineVisGUI, settings, GetPlugin()))
		return nullptr;
  
  CDialog *cdlg = dlg->Get();

  if (!cdlg) 
		return nullptr;
    
  return cdlg;
};

void SplineVisCustomGUI::Free(CDialog* dlg, void* userdata)
{
	if (!dlg || !userdata) 
		return;

  SplineVisGUI* sub = static_cast<SplineVisGUI*>(userdata);
  DeleteObj(sub);
};

const Char* SplineVisCustomGUI::GetResourceSym()
{
	// The symbol used in resource files
	return "SPLNEVISCUSTOMGUI";
};

CustomProperty* SplineVisCustomGUI::GetProperties()
{
	return nullptr;
};

Int32 SplineVisCustomGUI::GetResourceDataType(Int32*& table)
{
	// Returns the list of data types this GUI can work with
	table = g_stringtable; 
	return sizeof(g_stringtable)/sizeof(Int32);
};

//---------------------------
/// The datatype class
//---------------------------
Int32 SplineVisDataType::GetId()
{
  return ID_DTYPE_SPLINE_VIS;
}

CustomDataType* SplineVisDataType::AllocData()
{
	iSplineVisDataType* data = NewObjClear(iSplineVisDataType);
	return data;
};

void SplineVisDataType::FreeData(CustomDataType* data)
{
	iSplineVisDataType* d =  static_cast<iSplineVisDataType*>(data);
	DeleteObj(d);
}

Bool SplineVisDataType::CopyData(
	const CustomDataType* src, CustomDataType* dst, AliasTrans* aliastrans)
{
	// copy the data to the given target data

	const iSplineVisDataType* s =  static_cast<const iSplineVisDataType*>(src);
	iSplineVisDataType* d =  static_cast<iSplineVisDataType*>(dst);

	if (!s || !d)
		return false;

	d->_points.Flush();
	iferr (d->_points.CopyFrom(s->_points))
    return false;

	return true;
}

Int32 SplineVisDataType::Compare(const CustomDataType* d1, const CustomDataType* d2)
{
	// compare the values of the given elements
	// this is also used to determine if a parameter has changed

	const	iSplineVisDataType* const s = static_cast<const	iSplineVisDataType*>(d1);
	const	iSplineVisDataType* const d = static_cast<const	iSplineVisDataType*>(d2);

	if (!s || !d)
		return 0;

	// This just compares the number of points. A better implementation would be to also compare 
	// the values of the points

	const maxon::Int countd1 = s->_points.GetCount();
	const maxon::Int countd2 = d->_points.GetCount();

	if (countd1 == countd2) 
		return 0;
	if (countd1 < countd2) 
		return -1;
	if (countd1 > countd2) 
		return 1;

	return 0;
}

Bool SplineVisDataType::WriteData(const CustomDataType* t_d, HyperFile* hf)
{
	// write the data of this datatype to the given hyperfile

	const iSplineVisDataType* const d = static_cast<const iSplineVisDataType*>(t_d);

	// save number of points
	const maxon::Int length = d->_points.GetCount();
	hf->WriteInt64((Int64)length);

	// save points
	for (Int64 i = 0; i < length; ++i)
	{
		hf->WriteVector(d->_points[i]);
	}

	return true;
}

Bool SplineVisDataType::ReadData(CustomDataType* t_d, HyperFile* hf, Int32 level)
{
	// reads the data of this datatype from the given hyperfile

	iSplineVisDataType* const d = static_cast<iSplineVisDataType*>(t_d);

	if (level > 0)
	{
		// get number of points
		Int64 length = 0;
		if (hf->ReadInt64(&length))
		{
			// read points
			for (Int64 i = 0; i < length; ++i)
			{
				Vector vec;
				if (hf->ReadVector(&vec))
				{
					iferr (d->_points.Append(vec))
            return false;
				}
			}
		}		
	}
	return true;
}

const Char* SplineVisDataType::GetResourceSym()
{
	// this symbol can be used in resource files
	return "SPLINEVIS";
}

void SplineVisDataType::GetDefaultProperties(BaseContainer &data)
{
	// the default values of this datatype

	// use the custom GUI as default
	data.SetInt32(DESC_CUSTOMGUI, ID_GUI_SPLINE_VIS);
	data.SetInt32(DESC_ANIMATE, DESC_ANIMATE_ON);
}

//---------------------------
/// Register
//---------------------------
Bool RegisterSplineVisDatatypeCustomGUI()
{
	// register custom datatype
	if (!RegisterCustomDataTypePlugin(
		GeLoadString(IDS_DTYPE_SPLINE_VIS),
		CUSTOMDATATYPE_INFO_LOADSAVE |
		CUSTOMDATATYPE_INFO_TOGGLEDISPLAY |
		CUSTOMDATATYPE_INFO_HASSUBDESCRIPTION,
		NewObjClear(SplineVisDataType),
		1))
		return false;
	
	// dummy library for custom GUI
	static BaseCustomGuiLib myDotCustomGUI;

	ClearMem(&myDotCustomGUI, sizeof(myDotCustomGUI));
	FillBaseCustomGui(myDotCustomGUI);

	if (!InstallLibrary(ID_GUI_SPLINE_VIS, &myDotCustomGUI, 1000, sizeof(myDotCustomGUI)))
		return false;

	// register custom GUI
	if (!RegisterCustomGuiPlugin(
		GeLoadString(IDS_SPLINE_VIS), 0, NewObjClear(SplineVisCustomGUI)))
		return false;

	return true;
}